# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import empty, matmul, tensor, diag_embed
from torch.nn import Module, Embedding, Parameter
from torch.nn.functional import normalize
from torch.nn.init import xavier_uniform_

from torchkge.models import Model
from torchkge.utils import get_rank, get_mask, get_rolling_matrix
from torchkge.exceptions import WrongDimensionError


class RESCALModel(Module, Model):
    """Implementation of RESCAL model detailed in 2011 paper by Nickel et al..\
    In the original paper, optimization is done using Alternating Least Squares (ALS). Here we use\
    iterative gradient descent optimization.

    References
    ----------
    * Maximilian Nickel, Volker Tresp, and Hans-Peter Kriegel.
      A Three-way Model for Collective Learning on Multi-relational Data.
      In Proceedings of the 28th International Conference on Machine Learning, 2011.
      https://dl.acm.org/citation.cfm?id=3104584

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities
    number_entities: int
        Number of entities in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_matrices: torch Parameter, shape = (number_relations, ent_emb_dim, ent_emb_dim)
        Contains the matrices of the relations. It is initialized with Xavier uniform.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__()
        self.ent_emb_dim = ent_emb_dim
        self.number_entities = n_entities
        self.number_relations = n_relations

        # initialize embedding objects
        self.entity_embeddings = Embedding(self.number_entities, self.ent_emb_dim)
        self.relation_matrices = Parameter(xavier_uniform_(empty(size=(self.number_relations,
                                                                       self.ent_emb_dim,
                                                                       self.ent_emb_dim))))

        # fill the embedding weights with Xavier initialized values
        self.entity_embeddings.weight = Parameter(xavier_uniform_(
            empty(size=(self.number_entities, self.ent_emb_dim))))

        # normalize the embeddings
        self.normalize_parameters()

    def forward(self, heads, tails, negative_heads, negative_tails, relations):
        """Forward pass on the current batch.

        Parameters
        ----------
        heads: torch tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails: torch tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        negative_heads: torch tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's negatively sampled heads.
        negative_tails: torch tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's negatively sampled tails.
        relations: torch tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        golden_triplets: torch tensor, dtype = float, shape = (batch_size)
            Estimation of the true value that should be 1 (by matrix factorization).
        negative_triplets: torch tensor, dtype = float, shape = (batch_size)
            Estimation of the true value that should be 0 (by matrix factorization).

        """
        return self.scoring_function(heads, tails, relations), \
            self.scoring_function(negative_heads, negative_tails, relations)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = self.relation_matrices[rels_idx]

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)
        return product.view(heads_idx.shape[0])

    def normalize_parameters(self):
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities or candidates.
        tails: torch.Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: torch.Tensor, shape = (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype = float
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: torch.Tensor, shape = (b_size) or (b_size, self.number_entities), dtype = float
            Tensor containing the matrix products h.W.t^t for each sample of the batch.

        """
        b_size = len(heads)

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            heads = heads.view(b_size, 1, self.ent_emb_dim)
            tails = tails.view(b_size, self.ent_emb_dim, 1)
            return matmul(matmul(heads, rel_mat), tails).view(b_size)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            heads = heads.view(b_size, 1, self.ent_emb_dim)
            tails = tails.transpose(1, 2)
        else:
            tails = tails.view(b_size, self.ent_emb_dim, 1)

        return matmul(matmul(heads, rel_mat), tails).view(b_size, -1)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities.
        t_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities.
        r_mat: torch.Tensor, shape = (b_size, ent_emb_dim, ent_emb_dim), dtype = float
            Tensor containing matrices of current relations.
        candidates: torch.Tensor, shape = (b_size, number_entities, ent_emb_dim), dtype = float
            Tensor containing all entities as candidates for each sample of the batch.

        """
        b_size = h_idx.shape[0]

        candidates = self.entity_embeddings.weight.data
        candidates = candidates.view(1, self.number_entities, self.ent_emb_dim)
        candidates = candidates.expand(b_size, self.number_entities, self.ent_emb_dim)

        h_emb = self.entity_embeddings(h_idx)
        t_emb = self.entity_embeddings(t_idx)
        r_mat = self.relation_matrices[r_idx]

        return h_emb, t_emb, r_mat, candidates

    def compute_ranks(self, e_emb, candidates, r_mat, e_idx, r_idx, true_idx, dictionary, heads=1):
        """

        Parameters
        ----------
        e_emb: torch tensor, shape = (batch_size, rel_emb_dim), dtype = float
            Tensor containing current embeddings of entities.
        candidates: torch tensor, shape = (b_size, number_entities, ent_emb_dim), dtype = float
            Tensor containing projected embeddings of all entities.
        r_mat: torch.Tensor, shape = (b_size, ent_emb_dim, ent_emb_dim), dtype = float
            Tensor containing current matrices of relations.
        e_idx: torch tensor, shape = (batch_size), dtype = long
            Tensor containing the indices of entities.
        r_idx: torch tensor, shape = (batch_size), dtype = long
            Tensor containing the indices of relations.
        true_idx: torch tensor, shape = (batch_size), dtype = long
            Tensor containing the true entity for each sample.
        dictionary: default dict
            Dictionary of keys (int, int) and values list of ints giving all possible entities for\
            the (entity, relation) pair.
        heads: integer
            1 ou -1 (must be 1 if entities are heads and -1 if entities are tails). We test\
             dissimilarity between heads * entities + relations and heads * targets.


        Returns
        -------
        rank_true_entities: torch Tensor, shape = (b_size), dtype = int
            Tensor containing the rank of the true entities when ranking any entity based on\
            estimation of 1 or 0.
        filtered_rank_true_entities: torch Tensor, shape = (b_size), dtype = int
            Tensor containing the rank of the true entities when ranking only true false entities\
            based on estimation of 1 or 0.

        """
        current_batch_size, _ = e_emb.shape

        if heads == 1:
            scores = self.compute_product(e_emb, candidates, r_mat)
        else:
            scores = self.compute_product(candidates, e_emb, r_mat)

        # filter out the true negative samples by assigning negative score
        filt_scores = scores.clone()
        for i in range(current_batch_size):
            true_targets = dictionary[e_idx[i].item(), r_idx[i].item()].copy()
            if len(true_targets) == 1:
                continue
            true_targets.remove(true_idx[i].item())
            true_targets = tensor(list(true_targets)).long()
            filt_scores[i][true_targets] = float(-1)

        # from dissimilarities, extract the rank of the true entity.
        rank_true_entities = get_rank(scores, true_idx)
        filtered_rank_true_entities = get_rank(filt_scores, true_idx)

        return rank_true_entities, filtered_rank_true_entities

    def evaluate_candidates(self, h_idx, t_idx, r_idx, kg):
        h_emb, t_emb, r_mat, candidates = self.evaluation_helper(h_idx, t_idx, r_idx)

        rank_true_tails, filt_rank_true_tails = self.compute_ranks(h_emb,
                                                                   candidates,
                                                                   r_mat, h_idx, r_idx,
                                                                   t_idx,
                                                                   kg.dict_of_tails,
                                                                   heads=1)
        rank_true_heads, filt_rank_true_heads = self.compute_ranks(t_emb,
                                                                   candidates,
                                                                   r_mat, t_idx, r_idx,
                                                                   h_idx,
                                                                   kg.dict_of_heads,
                                                                   heads=-1)

        return rank_true_tails, filt_rank_true_tails, rank_true_heads, filt_rank_true_heads


class DistMultModel(RESCALModel):
    """Implementation of DistMult model detailed in 2014 paper by Yang et al..

    References
    ----------
    * Bishan Yang, Wen-tau Yih, Xiaodong He, Jianfeng Gao, and Li Deng.
      Embedding Entities and Relations for Learning and Inference in Knowledge Bases.
      arXiv :1412.6575 [cs], December 2014. arXiv : 1412.6575.
      https://arxiv.org/abs/1412.6575

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities
    number_entities: int
        Number of entities in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_vectors: torch Parameter, shape = (number_relations, ent_emb_dim)
        Contains the vectors to build diagonal matrices of the relations. It is initialized with
        Xavier uniform.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        del self.relation_matrices
        self.relation_vectors = Parameter(
            xavier_uniform_(empty(size=(self.number_relations, self.ent_emb_dim))))

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = diag_embed(self.relation_vectors[rels_idx])

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)

        return product.view(heads_idx.shape[0])

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities.
        t_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities.
        r_mat: torch.Tensor, shape = (b_size, ent_emb_dim, ent_emb_dim), dtype = float
            Tensor containing matrices of current relations.
        candidates: torch.Tensor, shape = (b_size, number_entities, ent_emb_dim), dtype = float
            Tensor containing all entities as candidates for each sample of the batch.

        """
        b_size = h_idx.shape[0]

        candidates = self.entity_embeddings.weight.data
        candidates = candidates.view(1, self.number_entities, self.ent_emb_dim)
        candidates = candidates.expand(b_size, self.number_entities, self.ent_emb_dim)

        h_emb = self.entity_embeddings(h_idx)
        t_emb = self.entity_embeddings(t_idx)
        r_mat = diag_embed(self.relation_vectors[r_idx])

        return h_emb, t_emb, r_mat, candidates


class HolEModel(RESCALModel):
    """Implementation of HolE model detailed in 2015 paper by Nickel et al..

        References
        ----------
        * Maximilian Nickel, Lorenzo Rosasco, and Tomaso Poggio.
          Holographic Embeddings of Knowledge Graphs.
          arXiv :1510.04935 [cs, stat], October 2015. arXiv : 1510.04935.
          https://arxiv.org/abs/1510.04935

        Parameters
        ----------
        ent_emb_dim: int
            Dimension of embedding space.
        n_entities: int
            Number of entities in the current data set.
        n_relations: int
            Number of relations in the current data set.

        Attributes
        ----------
        ent_emb_dim: int
            Dimension of the embedding of entities
        number_entities: int
            Number of entities in the current data set.
        entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
            Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
             normalized.
        relation_vectors: torch Parameter, shape = (number_relations, ent_emb_dim)
            Contains the vectors to build circular matrices of the relations. It is initialized
            with Xavier uniform.

        """
    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        del self.relation_matrices
        self.relation_vectors = Parameter(
            xavier_uniform_(empty(size=(self.number_relations, self.ent_emb_dim))))

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = get_rolling_matrix(self.relation_vectors[rels_idx])

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)

        return product.view(heads_idx.shape[0])

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size,), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities.
        t_emb: torch.Tensor, shape = (b_size, ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities.
        r_mat: torch.Tensor, shape = (b_size, ent_emb_dim, ent_emb_dim), dtype = float
            Tensor containing matrices of current relations.
        candidates: torch.Tensor, shape = (b_size, number_entities, ent_emb_dim), dtype = float
            Tensor containing all entities as candidates for each sample of the batch.

        """
        b_size = h_idx.shape[0]

        candidates = self.entity_embeddings.weight.data
        candidates = candidates.view(1, self.number_entities, self.ent_emb_dim)
        candidates = candidates.expand(b_size, self.number_entities, self.ent_emb_dim)

        h_emb = self.entity_embeddings(h_idx)
        t_emb = self.entity_embeddings(t_idx)
        r_mat = get_rolling_matrix(self.relation_vectors[r_idx])

        return h_emb, t_emb, r_mat, candidates

    def normalize_parameters(self):
        pass


class ComplExModel(DistMultModel):
    """Implementation of ComplEx model detailed in 2016 paper by Trouillon et al..

    References
    ----------
    * Théo Trouillon, Johannes Welbl, Sebastian Riedel, Éric Gaussier, and Guillaume Bouchard.
      Complex Embeddings for Simple Link Prediction.
      arXiv :1606.06357 [cs, stat], June 2016. arXiv : 1606.06357.
      https://arxiv.org/abs/1606.06357

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities
    number_entities: int
        Number of entities in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_vectors: torch Parameter, shape = (number_relations, ent_emb_dim)
        Contains the vectors to build block-diagonal matrices of the relations. It is initialized
        with Xavier uniform.
    smaller_dim: int
        Number of 2x2 matrices on the diagonals of relation-specific matrices.
    """
    def __init__(self, ent_emb_dim, n_entities, n_relations):
        try:
            assert ent_emb_dim % 2 == 0
        except AssertionError:
            raise WrongDimensionError('Embedding dimension should be pair.')

        super().__init__(ent_emb_dim, n_entities, n_relations)

        self.smaller_dim = int(self.ent_emb_dim / 2)

        self.real_mask = get_mask(self.ent_emb_dim, 0, self.smaller_dim)
        self.im_mask = get_mask(self.ent_emb_dim, self.smaller_dim, self.ent_emb_dim)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities or candidates.
        tails: torch.Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: torch.Tensor, shape = (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype = float
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: torch.Tensor, shape = (b_size, 1) or (b_size, self.number_entities), dtype = float
            Tensor containing the matrix products h^t.W.t for each sample of the batch.

        """
        b_size = len(heads)
        r_re = rel_mat[:, self.real_mask][:, :, self.real_mask]
        r_im = rel_mat[:, self.im_mask][:, :, self.im_mask]

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_re = tails[:, :, self.real_mask].transpose(2, 1)
            t_im = tails[:, :, self.im_mask].transpose(2, 1)

        else:
            h_re = heads[:, :, self.real_mask]
            h_im = heads[:, :, self.im_mask]

            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        re = matmul(h_re, matmul(r_re, t_re) + matmul(r_im, t_im)).view(b_size, -1)
        im = matmul(h_im, matmul(r_re, t_im) - matmul(r_im, t_re)).view(b_size, -1)

        return re + im

    def normalize_parameters(self):
        pass


class AnalogyModel(DistMultModel):
    """Implementation of ANALOGY model detailed in 2017 paper by Liu et al..\
    According to their remark in the implementation details, the number of scalars on\
    the diagonal of each relation-specific matrix is by default set to be half the embedding\
    dimension.

    References
    ----------
    * Hanxiao Liu, Yuexin Wu, and Yiming Yang.
      Analogical Inference for Multi-Relational Embeddings.
      arXiv :1705.02426 [cs], May 2017. arXiv : 1705.02426.
      https://arxiv.org/abs/1705.02426

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.
    scalar_share: float
        Share of the diagonal elements of the relation-specific matrices to be scalars. By default\
        it is set to half according to the original paper.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities
    number_entities: int
        Number of entities in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_vectors: torch Parameter, shape = (number_relations, ent_emb_dim)
        Contains the vectors to build almost-diagonal matrices of the relations. It is initialized
        with Xavier uniform.
    number_scalars: int
        Number of diagonal elements of the relation-specific matrices to be scalars. By default\
        it is set to half according to the original paper.
    smaller_dim: int
        Number of 2x2 matrices on the diagonals of relation-specific matrices.
    """

    def __init__(self, ent_emb_dim, n_entities, n_relations, scalar_share=0.5):
        try:
            assert ent_emb_dim % 2 == 0
        except AssertionError:
            raise WrongDimensionError('Embedding dimension should be pair.')

        super().__init__(ent_emb_dim, n_entities, n_relations)

        self.number_scalars = int(self.ent_emb_dim * scalar_share)

        if (self.ent_emb_dim - self.number_scalars) % 2 == 1:
            # the diagonal blocks are 2x2 so this dimension needs to be pair
            self.number_scalars -= 1
        self.smaller_dim = int((self.ent_emb_dim - self.number_scalars) / 2)

        self.scalar_mask = get_mask(self.ent_emb_dim, 0, self.number_scalars)
        self.real_mask = get_mask(self.ent_emb_dim, self.number_scalars,
                                  self.number_scalars + self.smaller_dim)
        self.im_mask = get_mask(self.ent_emb_dim, self.number_scalars + self.smaller_dim,
                                self.ent_emb_dim)

        assert (self.real_mask.sum() == self.im_mask.sum() == self.smaller_dim)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current head entities or candidates.
        tails: torch.Tensor, shape = (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: torch.Tensor, shape = (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype = float
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: torch.Tensor, shape = (b_size, 1) or (b_size, self.number_entities), dtype = float
            Tensor containing the matrix products h^t.W.t for each sample of the batch.

        """
        b_size = len(heads)
        r_scalar = rel_mat[:, self.scalar_mask][:, :, self.scalar_mask]
        r_re = rel_mat[:, self.real_mask][:, :, self.real_mask]
        r_im = rel_mat[:, self.im_mask][:, :, self.im_mask]

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            h_scalar = heads[:, self.scalar_mask].view(b_size, 1, self.number_scalars)
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_scalar = tails[:, self.scalar_mask].view(b_size, self.number_scalars, 1)
            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            h_scalar = heads[:, self.scalar_mask].view(b_size, 1, self.number_scalars)
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_scalar = tails[:, :, self.scalar_mask].transpose(2, 1)
            t_re = tails[:, :, self.real_mask].transpose(2, 1)
            t_im = tails[:, :, self.im_mask].transpose(2, 1)

        else:
            h_scalar = heads[:, :, self.scalar_mask]
            h_re = heads[:, :, self.real_mask]
            h_im = heads[:, :, self.im_mask]

            t_scalar = tails[:, self.scalar_mask].view(b_size, self.number_scalars, 1)
            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        scalar = matmul(matmul(h_scalar, r_scalar), t_scalar).view(b_size, -1)
        re = matmul(h_re, matmul(r_re, t_re) + matmul(r_im, t_im)).view(b_size, -1)
        im = matmul(h_im, matmul(r_re, t_im) - matmul(r_im, t_re)).view(b_size, -1)

        return scalar + re + im

    def normalize_parameters(self):
        """According to original paper, no normalization should be done on the parameters.

        """
        pass



from .Models import Model

from .TranslationModels import TransEModel
from .TranslationModels import TransHModel
from .TranslationModels import TransRModel
from .TranslationModels import TransDModel

from .BilinearModels import RESCALModel
from .BilinearModels import DistMultModel
from .BilinearModels import HolEModel
from .BilinearModels import ComplExModel
from .BilinearModels import AnalogyModel



from .data_preprocessing import get_dictionaries, get_bern_probs
from .dissimilarities import l1_dissimilarity, l2_dissimilarity
from .losses import MarginLoss, LogisticLoss, MSE
from .operations import get_rank, get_mask, get_rolling_matrix, init_embedding
from .negative_sampling import fill_in_dicts



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import empty, bincount, cat, topk, zeros
from torch.nn import Embedding, Parameter
from torch.nn.init import xavier_uniform_


def init_embedding(n_vectors, dim):
    """Create a torch.nn.Embedding object with `n_vectors` samples and `dim` dimensions.
    """
    entity_embeddings = Embedding(n_vectors, dim)
    entity_embeddings.weight = Parameter(xavier_uniform_(empty(size=(n_vectors, dim))))
    return entity_embeddings


def get_mask(length, start, end):
    """Create a mask of length `length` filled with 0s except between indices `start` (included)\
    and `end` (excluded).

    Parameters
    ----------
    length: int
    start: int
    end: int

    Returns
    -------
    mask: torch.Tensor, shape=(length), dtype=byte
        Mask of length `length` filled with 0s except between indices `start` (included)\
        and `end` (excluded).
    """
    mask = zeros(length)
    mask[[i for i in range(start, end)]] = 1
    return mask.byte()


def get_rolling_matrix(x):
    """

    Parameters
    ----------
    x: torch.Tensor, shape=(b_size, dim)

    Returns
    -------
    mat: torch.Tensor, shape=(b_size, dim, dim)
        Rolling matrix sur that mat[i,j] = x[i - j mod(dim)]
    """
    b_size, dim = x.shape
    x = x.view(b_size, 1, dim)
    return cat([x.roll(i, dims=2) for i in range(dim)], dim=1)


def get_rank(data, true, low_values=False):
    """

    Parameters
    ----------
    data: torch.Tensor, dtype = float, shape = (n_sample, dimensions)
    true: torch.Tensor, dtype = int, shape = (n_sample)
    low_values: bool
        if True, best rank is the lowest score else it is the highest

    Returns
    -------
    ranks: torch.Tensor, dtype = int, shape = (n_sample)
        data[ranks[i]] = true[i]
    """
    true_data = data.gather(1, true.long().view(-1, 1))

    if low_values:
        return (data <= true_data).sum(dim=1)
    else:
        return (data >= true_data).sum(dim=1)


def compute_weight(mask, k):
    """

    Parameters
    ----------
    mask
    k

    """
    weight = bincount(mask)[mask]
    weight[weight > k] = k
    return 1 / weight.float()


def pad_with_last_value(t, k):
    """Pad a tensor with its last value

    Parameters
    ----------
    t: torch.Tensor, shape = (n,m)
    k: integer

    Returns
    -------
    tensor of shape (n, m+i) where the n added columns are replicates of the last one of t.

    """
    n, m = t.shape
    tmp = t[:, -1].reshape(n, 1).expand(n, k)
    return cat((t, tmp), dim=1)


def concatenate_diff_sizes(a, b):
    """Concatenate 2D tensors of different shape by padding with last value the one with
    shortest second dimension.

    Parameters
    ----------
    a: torch.Tensor, shape = (n, m)
    b: torch.Tensor, shape = (k, l)

    Returns
    -------
    torch.Tensor of shape (n+k, max(m, l))
    """
    try:
        _, i = a.shape
        _, j = b.shape
        if i < j:
            return cat((pad_with_last_value(a, j - i), b), dim=0)
        elif j < i:
            return cat((a, pad_with_last_value(b, i - j)), dim=0)
        else:
            return cat((a, b), dim=0)
    except ValueError:
        return cat((a, b), dim=0)


def process_dissimilarities(dissimilarities, true, k_max):
    """Compute the rank of the true entities and the best candidates for each fact.

    Parameters
    ----------
    dissimilarities: torch.Tensor, dtype = float, shape = (batch_size, n_ent)
    true: torch.Tensor, dtype = long, shape = (batch_size),
        index of the the true entities for current relation
    k_max: integer
        Maximum value of the Hit@K measure.

    Returns
    -------
    rank_true_entities: torch.Tensor, dtype = long, shape = (batch_size)
        Rank of the true entity among all possible entities (ranked by decreasing dissimilarity)
    sorted_candidates: torch.Tensor, dtype = long, shape = (batch_size, k_max)
        Top k_max entity candidates in term of smaller dissimilarity(h+r, t).
    """
    # return the rank of the true value along with the sorted top k_max candidates
    _, sorted_candidates = topk(dissimilarities, k_max, dim=1, largest=False, sorted=True)
    rank_true_entities = get_rank(dissimilarities, true)
    return rank_true_entities, sorted_candidates



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""


def l1_dissimilarity(a, b):
    """

    Parameters
    ----------
    a: torch.Tensor, dtype = float, shape = (n_sample, dim)
    b: torch.Tensor, dtype = float, shape = (n_sample, dim)

    Returns
    -------
    dist: torch.Tensor, dtype = float, shape = (n_sample)
        Tensor of the row_wise L1 distance.

    """
    return (a-b).norm(p=1, dim=1)


def l2_dissimilarity(a, b):
    """

    Parameters
    ----------
    a: torch.Tensor, dtype = float, shape = (n_sample, dim)
    b: torch.Tensor, dtype = float, shape = (n_sample, dim)

    Returns
    -------
    dist: torch.Tensor, dtype = float, shape = (n_sample)
        Tensor of the row_wise squared L2 distance.

    """
    return (a-b).norm(p=2, dim=1)**2



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""


class Model(object):
    def __init__(self):
        pass

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        pass

    def normalize_parameters(self):
        pass

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        pass



from torchkge.models import Model

from sknetwork.embedding import GSVD


class GSVDModel(object):

    def __init__(self, biad, k, eps):
        self.biad = biad
        self.k = k
        self.eps = eps
        self.emb

    def embed(self):
        emb = GSVD(embedding_dimension=k, regularization=eps)
        emb.fit(biad)
        return emb.embedding_, emb.features_




# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import empty, matmul, eye, arange, tensor
from torch.nn import Module, Parameter
from torch.nn.functional import normalize
from torch.nn.init import xavier_uniform_
from torch.cuda import empty_cache

from torchkge.models import Model
from torchkge.utils import get_rank, init_embedding, l2_dissimilarity

from tqdm import tqdm


class TransEModel(Module, Model):
    """Implementation of TransE model detailed in 2013 paper by Bordes et al..

    References
    ----------
    * Antoine Bordes, Nicolas Usunier, Alberto Garcia-Duran, Jason Weston, and Oksana Yakhnenko.
      Translating Embeddings for Modeling Multi-relational Data.
      In Advances in Neural Information Processing Systems 26, pages 2787–2795, 2013.
      https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.
    dissimilarity: function
        Used to compute dissimilarities (e.g. L1 or L2 dissimilarities).

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    number_entities: int
        Number of entities in the current data set.
    number_relations: int
        Number of relations in the current data set.
    dissimilarity: function
        Used to compute dissimilarities (e.g. L1 or L2 dissimilarities).
    entity_embeddings: torch Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_embeddings: torch Embedding, shape = (number_relations, ent_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
         normalized.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations, dissimilarity, rel_emb_dim=None):
        super().__init__()

        self.ent_emb_dim = ent_emb_dim
        self.number_entities = n_entities
        self.number_relations = n_relations
        self.dissimilarity = dissimilarity

        if rel_emb_dim is None:
            rel_emb_dim = ent_emb_dim
        else:
            self.rel_emb_dim = rel_emb_dim

        # initialize embeddings
        self.entity_embeddings = init_embedding(self.number_entities, self.ent_emb_dim)
        self.relation_embeddings = init_embedding(self.number_relations, rel_emb_dim)

        # normalize parameters
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data,
                                                         p=2, dim=1)

    def forward(self, heads, tails, negative_heads, negative_tails, relations):
        """Forward pass on the current batch.

        Parameters
        ----------
        heads: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        negative_heads: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's negatively sampled heads.
        negative_tails: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's negatively sampled tails.
        relations: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        golden_triplets: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t for golden triplets.
        negative_triplets: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t for negatively
            sampled triplets.

        """
        golden_triplets = self.scoring_function(heads, tails, relations)
        negative_triplets = self.scoring_function(negative_heads, negative_tails, relations)

        return golden_triplets, negative_triplets

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t.

        """
        # recover relations embeddings
        rels_emb = self.relation_embeddings(rels_idx)

        # recover, project and normalize entity embeddings
        h_emb = self.recover_project_normalize(heads_idx, normalize_=True)
        t_emb = self.recover_project_normalize(tails_idx, normalize_=True)

        return - self.dissimilarity(h_emb + rels_emb, t_emb)

    def recover_project_normalize(self, ent_idx, normalize_=True):
        """

        Parameters
        ----------
        ent_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of entities
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: torch.Tensor, dtype = float, shape = (batch_size, ent_emb_dim)
            Embedded entities normalized.

        """
        # recover entity embeddings
        ent_emb = self.entity_embeddings(ent_idx)

        # normalize entity embeddings
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        return ent_emb

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: torch.Tensor, shape = (b_size, rel_emb_dim, n_entities), dtype = float
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing current relations embeddings.

        """
        # recover, project and normalize entity embeddings
        all_idx = arange(0, self.number_entities).long()

        if h_idx.is_cuda:
            all_idx = all_idx.cuda()
        proj_candidates = self.recover_project_normalize(all_idx, normalize_=False)

        proj_h_emb = proj_candidates[h_idx]
        proj_t_emb = proj_candidates[t_idx]
        r_emb = self.relation_embeddings(r_idx)

        b_size, emb_dim = proj_h_emb.shape
        proj_candidates = proj_candidates.transpose(0, 1)
        proj_candidates = proj_candidates.view(1, emb_dim, self.number_entities)
        proj_candidates = proj_candidates.expand(b_size, emb_dim, self.number_entities)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb

    def compute_ranks(self, proj_e_emb, proj_candidates,
                      r_emb, e_idx, r_idx, true_idx, dictionary, heads=1):
        """

        Parameters
        ----------
        proj_e_emb: torch.Tensor, shape = (batch_size, rel_emb_dim), dtype = float
            Tensor containing current projected embeddings of entities.
        proj_candidates: torch.Tensor, shape = (b_size, rel_emb_dim, n_entities), dtype = float
            Tensor containing projected embeddings of all entities.
        r_emb: torch.Tensor, shape = (batch_size, ent_emb_dim), dtype = float
            Tensor containing current embeddings of relations.
        e_idx: torch.Tensor, shape = (batch_size), dtype = long
            Tensor containing the indices of entities.
        r_idx: torch.Tensor, shape = (batch_size), dtype = long
            Tensor containing the indices of relations.
        true_idx: torch.Tensor, shape = (batch_size), dtype = long
            Tensor containing the true entity for each sample.
        dictionary: default dict
            Dictionary of keys (int, int) and values list of ints giving all possible entities for
            the (entity, relation) pair.
        heads: integer
            1 ou -1 (must be 1 if entities are heads and -1 if entities are tails). \
            We test dissimilarity between heads * entities + relations and heads * targets.


        Returns
        -------
        rank_true_entities: torch.Tensor, shape = (b_size), dtype = int
            Tensor containing the rank of the true entities when ranking any entity based on \
            computation of d(hear+relation, tail).
        filtered_rank_true_entities: torch.Tensor, shape = (b_size), dtype = int
            Tensor containing the rank of the true entities when ranking only true false entities \
            based on computation of d(hear+relation, tail).

        """
        current_batch_size, embedding_dimension = proj_e_emb.shape

        # tmp_sum is either heads + r_emb or r_emb - tails (expand does not use extra memory)
        tmp_sum = (heads * proj_e_emb + r_emb).view((current_batch_size, embedding_dimension, 1))
        tmp_sum = tmp_sum.expand((current_batch_size, embedding_dimension, self.number_entities))

        # compute either dissimilarity(heads + relation, proj_candidates) or
        # dissimilarity(-proj_candidates, relation - tails)
        dissimilarities = self.dissimilarity(tmp_sum, heads * proj_candidates)

        # filter out the true negative samples by assigning infinite dissimilarity
        filt_dissimilarities = dissimilarities.clone()
        for i in range(current_batch_size):
            true_targets = dictionary[e_idx[i].item(), r_idx[i].item()].copy()
            if len(true_targets) == 1:
                continue
            true_targets.remove(true_idx[i].item())
            true_targets = tensor(list(true_targets)).long()
            filt_dissimilarities[i][true_targets] = float('Inf')

        # from dissimilarities, extract the rank of the true entity.
        rank_true_entities = get_rank(-dissimilarities, true_idx)
        filtered_rank_true_entities = get_rank(-filt_dissimilarities, true_idx)

        return rank_true_entities, filtered_rank_true_entities

    def evaluate_candidates(self, h_idx, t_idx, r_idx, kg):
        proj_h_emb, proj_t_emb, proj_candidates, r_emb = self.evaluation_helper(h_idx, t_idx, r_idx)

        # evaluation_helper both ways (head, rel) -> tail and (rel, tail) -> head
        rank_true_tails, filt_rank_true_tails = self.compute_ranks(proj_h_emb,
                                                                   proj_candidates,
                                                                   r_emb, h_idx, r_idx,
                                                                   t_idx,
                                                                   kg.dict_of_tails,
                                                                   heads=1)
        rank_true_heads, filt_rank_true_heads = self.compute_ranks(proj_t_emb,
                                                                   proj_candidates,
                                                                   r_emb, t_idx, r_idx,
                                                                   h_idx,
                                                                   kg.dict_of_heads,
                                                                   heads=-1)

        return rank_true_tails, filt_rank_true_tails, rank_true_heads, filt_rank_true_heads


class TransHModel(TransEModel):
    """Implementation of TransH model detailed in 2014 paper by Wang et al..

    References
    ----------
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      Knowledge Graph Embedding by Translating on Hyperplanes.
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.
      https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    number_entities: int
        Number of entities in the current data set.
    number_relations: int
        Number of relations in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_embeddings: torch.nn.Embedding, shape = (number_relations, ent_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
        normalized.
    dissimilarity: function
        `torchkge.utils.dissimilarities.l2_dissimilarity`
    normal_vectors: torch.Tensor, shape = (number_relations, ent_emb_dim)
        Normal vectors associated to each relation and used to compute the relation-specific\
        hyperplanes entities are projected on. See paper for more details.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):

        super().__init__(ent_emb_dim, n_entities, n_relations, l2_dissimilarity)

        self.normal_vectors = Parameter(xavier_uniform_(empty(size=(n_relations, ent_emb_dim))))
        self.normalize_parameters()

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        # recover relations embeddings and normal projection vectors
        relations_embeddings = self.relation_embeddings(rels_idx)
        normal_vectors = normalize(self.normal_vectors[rels_idx], p=2, dim=1)

        # project entities in relation specific hyperplane
        projected_heads = self.recover_project_normalize(heads_idx, normalize_=True,
                                                         normal_vectors=normal_vectors)
        projected_tails = self.recover_project_normalize(tails_idx, normalize_=True,
                                                         normal_vectors=normal_vectors)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, normalize_=True, **kwargs):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided normal vectors.

        Parameters
        ----------
        ent_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of entities
        normalize_: bool
            Whether entities embeddings should be normalized or not.
        normal_vectors: torch.Tensor, dtype = float, shape = (batch_size, ent_emb_dim)
            Normal vectors relative to the current relations.

        Returns
        -------
        projections: torch.Tensor, dtype = float, shape = (batch_size, ent_emb_dim)
            Projection of the embedded entities on the hyperplanes defined by the provided normal\
            vectors.

        """
        # recover entity embeddings
        ent_emb = self.entity_embeddings(ent_idx)

        # normalize entity embeddings
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        # project entities into relation space
        normal_vectors = kwargs['normal_vectors']
        normal_component = (ent_emb * normal_vectors).sum(dim=1).view((-1, 1))

        return ent_emb - normal_component * normal_vectors

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.normal_vectors.data = normalize(self.normal_vectors, p=2, dim=1)
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: torch.Tensor, shape = (b_size, rel_emb_dim, n_entities), dtype = float
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing current relations embeddings.

        """
        # recover relations embeddings and normal projection vectors
        r_emb = self.relation_embeddings(r_idx)
        normal_vectors = normalize(self.normal_vectors[r_idx], p=2, dim=1)
        b_size, _ = normal_vectors.shape

        # recover candidates
        all_idx = arange(0, self.number_entities).long()
        if h_idx.is_cuda:
            all_idx = all_idx.cuda()
        candidates = self.entity_embeddings(all_idx).transpose(0, 1)
        candidates = candidates.view((1,
                                      self.ent_emb_dim,
                                      self.number_entities)).expand((b_size,
                                                                     self.ent_emb_dim,
                                                                     self.number_entities))

        # project each candidates with each normal vector
        normal_components = candidates * normal_vectors.view((b_size, self.ent_emb_dim, 1))
        normal_components = normal_components.sum(dim=1).view(b_size, 1, self.number_entities)
        normal_components = normal_components * normal_vectors.view(b_size, self.ent_emb_dim, 1)
        proj_candidates = candidates - normal_components

        assert proj_candidates.shape == (b_size, self.ent_emb_dim, self.number_entities)

        # recover, project and normalize entity embeddings
        mask = h_idx.view(b_size, 1, 1).expand(b_size, self.ent_emb_dim, 1)
        proj_h_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.ent_emb_dim)

        mask = t_idx.view(b_size, 1, 1).expand(b_size, self.ent_emb_dim, 1)
        proj_t_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.ent_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TransRModel(TransEModel):
    """Implementation of TransR model detailed in 2015 paper by Lin et al..

    References
    ----------
    * Yankai Lin, Zhiyuan Liu, Maosong Sun, Yang Liu, and Xuan Zhu.
      Learning Entity and Relation Embeddings for Knowledge Graph Completion.
      In Twenty-Ninth AAAI Conference on Artificial Intelligence, February 2015
      https://www.aaai.org/ocs/index.php/AAAI/AAAI15/paper/view/9571/9523

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    number_entities: int
        Number of entities in the current data set.
    number_relations: int
        Number of relations in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_embeddings: torch.nn.Embedding, shape = (number_relations, rel_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
        normalized.
    dissimilarity: function
        `torchkge.utils.dissimilarities.l2_dissimilarity`
    projection_matrices: torch.Tensor, shape = (number_relations, rel_emb_dim, ent_emb_dim)
        Relation-specific projection matrices. See paper for more details.

    """

    def __init__(self, ent_emb_dim, rel_emb_dim, n_entities, n_relations):

        super().__init__(ent_emb_dim, n_entities, n_relations, l2_dissimilarity,
                         rel_emb_dim=rel_emb_dim)

        self.projection_matrices = Parameter(xavier_uniform_(empty(size=(n_relations, rel_emb_dim,
                                                                         ent_emb_dim))))
        self.normalize_parameters()

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        # recover relations embeddings and normal projection matrices
        relations_embeddings = normalize(self.relation_embeddings(rels_idx), p=2, dim=1)

        # projection_matrices = normalize(self.projection_matrices[rels_idx], p=2, dim=2)
        projection_matrices = self.projection_matrices[rels_idx]

        # project entities in relation specific hyperplane
        projected_heads = self.recover_project_normalize(heads_idx, normalize_=True,
                                                         projection_matrices=projection_matrices)
        projected_tails = self.recover_project_normalize(tails_idx, normalize_=True,
                                                         projection_matrices=projection_matrices)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, normalize_=True, **kwargs):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided projection matrices.

        Parameters
        ----------
        ent_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of entities
        normalize_: bool
            Whether entities embeddings should be normalized or not.
        projection_matrices: torch.Tensor, dtype = float, shape = (b_size, r_emb_dim, e_emb_dim)
            Projection matrices for the current relations.

        Returns
        -------
        projections: torch.Tensor, dtype = float, shape = (batch_size, rel_emb_dim)
            Projection of the entities into relation-specific subspaces.
        """
        b_size = ent_idx.shape[0]
        # recover and normalize embeddings
        ent_emb = self.entity_embeddings(ent_idx)
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        # project entities into relation space
        new_shape = (b_size, self.ent_emb_dim, 1)
        projection_matrices = kwargs['projection_matrices']
        projection = matmul(projection_matrices, ent_emb.view(new_shape))

        return normalize(projection.view(b_size, self.rel_emb_dim), p=2, dim=1)

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data,
                                                         p=2, dim=1)
        self.projection_matrices.data = normalize(self.projection_matrices.data, p=2, dim=2)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: torch.Tensor, shape = (b_size, rel_emb_dim, n_entities), dtype = float
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: torch.Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing current relations embeddings.

        """
        # recover relations embeddings and normal projection matrices
        r_emb = normalize(self.relation_embeddings(r_idx), p=2, dim=1)
        projection_matrices = normalize(self.projection_matrices[r_idx], p=2, dim=2)
        b_size, _, _ = projection_matrices.shape

        # recover candidates
        all_idx = arange(0, self.number_entities).long()
        if h_idx.is_cuda:
            all_idx = all_idx.cuda()
        candidates = self.entity_embeddings(all_idx).transpose(0, 1)
        candidates = candidates.view((1,
                                      self.ent_emb_dim,
                                      self.number_entities)).expand((b_size,
                                                                     self.ent_emb_dim,
                                                                     self.number_entities))

        # project each candidates with each projection matrix
        proj_candidates = matmul(projection_matrices, candidates)

        mask = h_idx.view(b_size, 1, 1).expand(b_size, self.rel_emb_dim, 1)
        proj_h_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.rel_emb_dim)

        mask = t_idx.view(b_size, 1, 1).expand(b_size, self.rel_emb_dim, 1)
        proj_t_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.rel_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TransDModel(TransEModel):
    """Implementation of TransD model detailed in 2015 paper by Ji et al..

    References
    ----------
    * Guoliang Ji, Shizhu He, Liheng Xu, Kang Liu, and Jun Zhao.
      Knowledge Graph Embedding via Dynamic Mapping Matrix.
      In Proceedings of the 53rd Annual Meeting of the Association for Computational Linguistics and
      the 7th International Joint Conference on Natural Language Processing (Volume 1: Long Papers)
      pages 687–696, Beijing, China, July 2015. Association for Computational Linguistics.
      https://aclweb.org/anthology/papers/P/P15/P15-1067/

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    number_entities: int
        Number of entities in the current data set.
    number_relations: int
        Number of relations in the current data set.
    entity_embeddings: torch.nn.Embedding, shape = (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_embeddings: torch.nn.Embedding, shape = (number_relations, rel_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
        normalized.
    dissimilarity: function
        `torchkge.utils.dissimilarities.l2_dissimilarity`
    ent_proj_vects: torch.Tensor, shape = (number_entities, ent_emb_dim)
        Entity-specific vector used to build projection matrices. See paper for more details.
    rel_proj_vects: torch.Tensor, shape = (number_relations, rel_emb_dim)
        Relation-specific vector used to build projection matrices. See paper for more details.

    """

    def __init__(self, ent_emb_dim, rel_emb_dim, n_entities, n_relations):

        super().__init__(ent_emb_dim, n_entities, n_relations, l2_dissimilarity,
                         rel_emb_dim=rel_emb_dim)

        self.ent_proj_vects = Parameter(xavier_uniform_(empty(size=(n_entities, ent_emb_dim))))
        self.rel_proj_vects = Parameter(xavier_uniform_(empty(size=(n_relations, rel_emb_dim))))
        self.normalize_parameters()

        self.evaluated_projections = False
        self.projected_entities = Parameter(empty(size=(self.number_relations,
                                                        self.rel_emb_dim,
                                                        self.number_entities)), requires_grad=False)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's heads
        tails_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: torch.Tensor, dtype = float, shape = (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        self.evaluated_projections = False

        # recover relations projection vectors and relations embeddings
        rel_proj = normalize(self.rel_proj_vects[rels_idx], p=2, dim=1)
        relations_embeddings = normalize(self.relation_embeddings(rels_idx), p=2, dim=1)

        # project
        projected_heads = self.recover_project_normalize(heads_idx, normalize_=True,
                                                         rel_proj=rel_proj)
        projected_tails = self.recover_project_normalize(tails_idx, normalize_=True,
                                                         rel_proj=rel_proj)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, normalize_=True, **kwargs):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided normal vectors.

        Parameters
        ----------
        ent_idx: torch.Tensor, dtype = long, shape = (batch_size)
            Integer keys of entities
        normalize_: bool
            Whether entities embeddings should be normalized or not.
        rel_proj: torch.Tensor, dtype = float, shape = (batch_size, rel_emb_dim)
            Projection vectors for the current relations.

        Returns
        -------
        projections: torch.Tensor, dtype = float, shape = (batch_size, rel_emb_dim)
            Projection of the entities into relation-specific subspaces.

        """
        b_size = ent_idx.shape[0]

        # recover entities embeddings and projection vectors
        ent_emb = self.entity_embeddings(ent_idx)
        ent_proj = self.ent_proj_vects[ent_idx]

        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)
            ent_proj = normalize(ent_proj, p=2, dim=1)

        # project entities into relation space
        rel_proj = kwargs['rel_proj']
        proj_mat = matmul(rel_proj.view((b_size, self.rel_emb_dim, 1)),
                          ent_proj.view((b_size, 1, self.ent_emb_dim)))

        if proj_mat.is_cuda:
            proj_mat += eye(n=self.rel_emb_dim, m=self.ent_emb_dim, device='cuda')
        else:
            proj_mat += eye(n=self.rel_emb_dim, m=self.ent_emb_dim)

        projection = matmul(proj_mat, ent_emb.view((b_size, self.ent_emb_dim, 1)))

        return projection.view((b_size, self.rel_emb_dim))

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data,
                                                         p=2, dim=1)
        self.ent_proj_vects.data = normalize(self.ent_proj_vects.data, p=2, dim=1)
        self.rel_proj_vects.data = normalize(self.rel_proj_vects.data, p=2, dim=1)

    def evaluate_projections(self):
        """Project all entities according to each relation.
        """
        # TODO turn this to batch computation

        if self.evaluated_projections:
            return

        print('Projecting entities in relations spaces.')

        for i in tqdm(range(self.number_entities)):
            ent_proj_vect = self.ent_proj_vects.data[i].view(1, -1)
            rel_proj_vects = self.rel_proj_vects.data.view(self.number_relations,
                                                           self.rel_emb_dim, 1)

            projection_matrices = matmul(rel_proj_vects, ent_proj_vect)

            if projection_matrices.is_cuda:
                id_mat = eye(n=self.rel_emb_dim, m=self.ent_emb_dim, device='cuda')
            else:
                id_mat = eye(n=self.rel_emb_dim, m=self.ent_emb_dim)

            id_mat = id_mat.view(1, self.rel_emb_dim, self.ent_emb_dim)

            projection_matrices += id_mat.expand(self.number_relations, self.rel_emb_dim,
                                                 self.ent_emb_dim)

            empty_cache()

            mask = tensor([i]).long()

            if self.entity_embeddings.weight.is_cuda:
                assert self.projected_entities.is_cuda
                empty_cache()
                mask = mask.cuda()

            entity = self.entity_embeddings(mask)
            projected_entity = matmul(projection_matrices, entity.view(-1)).detach()
            projected_entity = projected_entity.view(self.number_relations, self.rel_emb_dim, 1)
            self.projected_entities[:, :, i] = projected_entity.view(self.number_relations,
                                                                     self.rel_emb_dim)

            del projected_entity

        self.evaluated_projections = True

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current head entities.
        t_idx: torch.Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current tail entities.
        r_idx: torch Tensor, shape = (b_size), dtype = long
            Tensor containing indices of current relations.
        Returns
        -------
        proj_h_emb: torch Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: torch Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: torch Tensor, shape = (b_size, rel_emb_dim, n_entities), dtype = float
            Tensor containing all entities projected in each relation spaces (relations\
            corresponding to current batch's relations).
        r_emb: torch Tensor, shape = (b_size, rel_emb_dim), dtype = float
            Tensor containing current relations embeddings.

        """
        b_size = h_idx.shape[0]
        if not self.evaluated_projections:
            self.evaluate_projections()

        # recover relations embeddings and projected candidates
        r_emb = normalize(self.relation_embeddings(r_idx), p=2, dim=1)
        proj_candidates = self.projected_entities[r_idx]

        mask = h_idx.view(b_size, 1, 1).expand(b_size, self.rel_emb_dim, 1)
        proj_h_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.rel_emb_dim)

        mask = t_idx.view(b_size, 1, 1).expand(b_size, self.rel_emb_dim, 1)
        proj_t_emb = proj_candidates.gather(dim=2, index=mask).view(b_size, self.rel_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""


class NotYetEvaluatedError(Exception):
    def __init__(self, message):
        super().__init__(message)


class SizeMismatchError(Exception):
    def __init__(self, message):
        super().__init__(message)


class WrongDimensionError(Exception):
    def __init__(self, message):
        super().__init__(message)


class NotYetImplementedError(Exception):
    def __init__(self, message):
        super().__init__(message)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch.nn import Module, MarginRankingLoss, SoftMarginLoss
from torch import ones_like


class MarginLoss(Module):
    def __init__(self, margin):
        super().__init__()
        self.loss = MarginRankingLoss(margin=margin, reduction='sum')

    def forward(self, output):
        golden_triplets, negative_triplets = output[0], output[1]
        return self.loss(golden_triplets, negative_triplets, target=ones_like(golden_triplets))


class LogisticLoss(Module):
    def __init__(self):
        super().__init__()
        self.loss = SoftMarginLoss(reduction='sum')

    def forward(self, output):
        golden_triplets, negative_triplets = output[0], output[1]
        targets = ones_like(golden_triplets)
        return self.loss(golden_triplets, targets) + self.loss(negative_triplets, -targets)


class MSE(Module):
    def __init__(self):
        super().__init__()

    def forward(self, output):
        golden_triplets, negative_triplets = output[0], output[1]
        return 1/2 * (((1 - golden_triplets)**2).sum() + (negative_triplets**2).sum())



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import empty
from torch.utils.data import DataLoader
from torchkge.exceptions import NotYetEvaluatedError

from tqdm import tqdm


class LinkPredictionEvaluator(object):
    """Evaluate performance of given embedding using link prediction method.

    Parameters
    ----------
    model: torchkge model
    knowledge_graph: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.

    Attributes
    ----------
    model: torchkge.models.Model
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.
    rank_true_heads: torch tensor, shape = (n_facts), dtype = int
        Rank of the true head when all possible entities are ranked in term of dissimilarity\
        with tail - relation.
    rank_true_tails: torch tensor, shape = (n_facts), dtype = int
        Rank of the true tail when all possible entities are ranked in term of dissimilarity\
        with head + relation.
    filt_rank_true_heads: torch tensor, shape = (n_facts), dtype = int
        Filtered rank of the true tail when all possible entities are ranked in term of\
        dissimilarity with head + relation.
    filt_rank_true_tails: torch tensor, shape = (n_facts), dtype = int
        Filtered rank of the true tail when all possible entities are ranked in term of\
        dissimilarity with head + relation.
    evaluated: bool
        Indicates if the method LinkPredictionEvaluator.evaluate() has been called on\
        current object
    k_max: bool, default = 10
        Max value to be used to compute the hit@k score.

    References
    ----------
    * Antoine Bordes, Nicolas Usunier, Alberto Garcia-Duran, Jason Weston, and Oksana Yakhnenko.
      Translating Embeddings for Modeling Multi-relational Data.
      In Advances in Neural Information Processing Systems 26, pages 2787–2795, 2013.
      https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data

    """

    def __init__(self, model, knowledge_graph):
        self.model = model
        self.kg = knowledge_graph

        self.rank_true_heads = empty(size=(knowledge_graph.n_sample,)).long()
        self.rank_true_tails = empty(size=(knowledge_graph.n_sample,)).long()
        self.filt_rank_true_heads = empty(size=(knowledge_graph.n_sample,)).long()
        self.filt_rank_true_tails = empty(size=(knowledge_graph.n_sample,)).long()

        self.evaluated = False
        self.k_max = 10

    def evaluate(self, batch_size, k_max):
        """

        Parameters
        ----------
        batch_size: int
            Size of the current batch.
        k_max: int
            Maximal k value we plan to use for Hit@k. This is used to truncate tensor so that it\
            fits in memory.

        """
        self.k_max = k_max
        use_cuda = self.model.entity_embeddings.weight.is_cuda
        dataloader = DataLoader(self.kg, batch_size=batch_size, pin_memory=use_cuda)

        if use_cuda:
            self.rank_true_heads = self.rank_true_heads.cuda()
            self.rank_true_tails = self.rank_true_tails.cuda()
            self.filt_rank_true_heads = self.filt_rank_true_heads.cuda()
            self.filt_rank_true_tails = self.filt_rank_true_tails.cuda()

        for i, batch in tqdm(enumerate(dataloader), total=len(dataloader)):

            h_idx, t_idx, r_idx = batch[0], batch[1], batch[2]

            if h_idx.is_pinned():
                h_idx, t_idx, r_idx = h_idx.cuda(), t_idx.cuda(), r_idx.cuda()

            rank_true_tails, filt_rank_true_tails, rank_true_heads, filt_rank_true_heads \
                = self.model.evaluate_candidates(h_idx, t_idx, r_idx, self.kg)

            self.rank_true_heads[i * batch_size: (i + 1) * batch_size] = rank_true_heads
            self.rank_true_tails[i * batch_size: (i + 1) * batch_size] = rank_true_tails

            self.filt_rank_true_heads[i * batch_size: (i + 1) * batch_size] = filt_rank_true_heads
            self.filt_rank_true_tails[i * batch_size: (i + 1) * batch_size] = filt_rank_true_tails

        self.evaluated = True

        if use_cuda:
            self.rank_true_heads = self.rank_true_heads.cpu()
            self.rank_true_tails = self.rank_true_tails.cpu()
            self.filt_rank_true_heads = self.filt_rank_true_heads.cpu()
            self.filt_rank_true_tails = self.filt_rank_true_tails.cpu()

    def mean_rank(self):
        """

        Returns
        -------
        mean_rank: float
            Mean rank of the true entity when replacing alternatively head and tail in\
            any fact of the dataset.
        filt_mean_rank: float
            Filtered mean rank of the true entity when replacing alternatively head and tail in\
            any fact of the dataset.

        """
        if not self.evaluated:
            raise NotYetEvaluatedError('Evaluator not evaluated call '
                                       'LinkPredictionEvaluator.evaluate')
        sum_ = (self.rank_true_heads.float().mean() +
                self.rank_true_tails.float().mean()).item()
        filt_sum = (self.filt_rank_true_heads.float().mean() +
                    self.filt_rank_true_tails.float().mean()).item()
        return sum_ / 2, filt_sum / 2

    def hit_at_k_heads(self, k=10):
        if not self.evaluated:
            raise NotYetEvaluatedError('Evaluator not evaluated call '
                                       'LinkPredictionEvaluator.evaluate')
        head_hit = (self.rank_true_heads < k).float().mean()
        filt_head_hit = (self.filt_rank_true_heads < k).float().mean()

        return head_hit.item(), filt_head_hit.item()

    def hit_at_k_tails(self, k=10):
        if not self.evaluated:
            raise NotYetEvaluatedError('Evaluator not evaluated call '
                                       'LinkPredictionEvaluator.evaluate')
        tail_hit = (self.rank_true_tails < k).float().mean()
        filt_tail_hit = (self.filt_rank_true_tails < k).float().mean()

        return tail_hit.item(), filt_tail_hit.item()

    def hit_at_k(self, k=10):
        """

        Parameters
        ----------
        k: int
            Hit@k is the number of entities that show up in the top k that give facts present\
            in the dataset.

        Returns
        -------
        avg_hitatk: float
            Average of hit@k for head and tail replacement. Computation is done in a\
            vectorized way.
        filt_avg_hitatk: float
            Filtered Average of hit@k for head and tail replacement. Computation is done in a\
            vectorized way.

        """
        if not self.evaluated:
            raise NotYetEvaluatedError('Evaluator not evaluated call '
                                       'LinkPredictionEvaluator.evaluate')

        head_hit, filt_head_hit = self.hit_at_k_heads(k=k)
        tail_hit, filt_tail_hit = self.hit_at_k_tails(k=k)

        return (head_hit + tail_hit) / 2, (filt_head_hit + filt_tail_hit) / 2

    def mrr(self):
        """

        Returns
        -------
        avg_mrr: float
            Average of mean recovery rank for head and tail replacement. Computation is done in a\
            vectorized way.
        filt_avg_mrr: float
            Filtered Average of mean recovery rank for head and tail replacement. Computation is\
            done in a vectorized way.

        """
        if not self.evaluated:
            raise NotYetEvaluatedError('Evaluator not evaluated call '
                                       'LinkPredictionEvaluator.evaluate')
        head_mrr = (self.rank_true_heads.float()**(-1)).mean()
        tail_mrr = (self.rank_true_tails.float()**(-1)).mean()
        filt_head_mrr = (self.filt_rank_true_heads.float()**(-1)).mean()
        filt_tail_mrr = (self.filt_rank_true_tails.float()**(-1)).mean()

        return (head_mrr + tail_mrr).item() / 2, (filt_head_mrr + filt_tail_mrr).item() / 2



# -*- coding: utf-8 -*-

"""Top-level package for TorchKGE."""

__author__ = """Armand Boschin"""
__email__ = 'aboschin@enst.fr'
__version__ = '0.10.6'

from .data import KnowledgeGraph

from .models import TransEModel, TransHModel, TransRModel, TransDModel
from .models import RESCALModel, DistMultModel

from torchkge.utils import l1_dissimilarity, l2_dissimilarity
from torchkge.utils import MarginLoss, LogisticLoss, MSE

from .evaluation.LinkPrediction import LinkPredictionEvaluator

from torchkge.exceptions import NotYetEvaluatedError



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""
from collections import defaultdict
from tqdm import tqdm


def fill_in_dicts(kg, possible_heads=None, possible_tails=None):
    if possible_heads is None:
        possible_heads = defaultdict(set)
    if possible_tails is None:
        possible_tails = defaultdict(set)

    for i in tqdm(range(kg.n_sample)):
        possible_heads[kg.relations[i].item()].add(kg.head_idx[i].item())
        possible_tails[kg.relations[i].item()].add(kg.tail_idx[i].item())

    return possible_heads, possible_tails



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""
from torch import zeros, cat
from torch.utils.data import DataLoader
from torchkge.data import SmallKG

from torchkge.sampling import PositionalNegativeSampler


class TripletClassificationEvaluator(object):
    """Evaluate performance of given embedding using triplet classification method.

    References
    ----------
    * Richard Socher, Danqi Chen, Christopher D Manning, and Andrew Ng.
      Reasoning With Neural Tensor Networks for Knowledge Base Completion.
      In Advances in Neural Information Processing Systems 26, pages 926–934. 2013.
      https://nlp.stanford.edu/pubs/SocherChenManningNg_NIPS2013.pdf

    Parameters
    ----------
    model: torchkge.models.Model
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.

    Attributes
    ----------
    model: torchkge.models.Model
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Knowledge graph in the form of an object implemented in torchkge.data.KnowledgeGraph.
    use_cuda: bool
        Indicate whether to use cuda or not. It is defined by looking into the device of parameters\
         of the model.
    evaluated: bool
        Indicate whether the `evaluate` function has been called.
    thresholds: float
        Float value of the thresholds for the scoring function to consider a triplet as true. It is\
        defined by calling the `evaluate` function.
    sampler: torchkge.sampling.NegativeSampler
        Negative sampler.

    """

    def __init__(self, model, kg_val, kg_test):
        self.model = model
        self.kg_val = kg_val
        self.kg_test = kg_test
        self.use_cuda = self.model.entity_embeddings.weight.is_cuda

        self.evaluated = False
        self.thresholds = None

        self.sampler = PositionalNegativeSampler(self.kg_val, kg_test=self.kg_test)

    def get_scores(self, heads, tails, relations, batch_size):
        """With head, tail and relation indexes, compute the value of the scoring function of the\
        model.

        Parameters
        ----------
        heads: torch.Tensor, dtype = long, shape = n_sample
            List of heads indices.
        tails: torch.Tensor, dtype = long, shape = n_sample
            List of tails indices.
        relations: torch.Tensor, dtype = long, shape = n_sample
            List of relation indices.
        batch_size: int

        Returns
        -------
        scores: torch.Tensor, dtype = float, shape = n_sample
            List of scores of each triplet.
        """
        scores = []

        small_kg = SmallKG(heads, tails, relations)
        dataloader = DataLoader(small_kg, batch_size=batch_size, pin_memory=self.use_cuda)
        for i, batch in enumerate(dataloader):

            h_idx, t_idx, r_idx = batch[0], batch[1], batch[2]

            if h_idx.is_pinned():
                h_idx, t_idx, r_idx = h_idx.cuda(), t_idx.cuda(), r_idx.cuda()

            scores.append(self.model.scoring_function(h_idx, t_idx, r_idx))

        return cat(scores, dim=0)

    def evaluate(self, batch_size):
        """Find relation thresholds using the validation set. As described in the paper by\
         Socher et al., for a relation, the threshold is a value t such that if the score of a\
         triplet is larger than t, the fact is true.

        Parameters
        ----------
        batch_size: int

        """
        r_idx = self.kg_val.relations

        neg_heads, neg_tails = self.sampler.corrupt_kg(batch_size, self.use_cuda, which='main')
        neg_scores = self.get_scores(neg_heads, neg_tails, r_idx, batch_size)

        self.thresholds = zeros(self.kg_val.n_rel)

        for i in range(self.kg_val.n_rel):
            mask = (r_idx == i).byte()
            assert mask.sum() > 0
            self.thresholds[i] = neg_scores[mask == 1].max()

        self.evaluated = True
        self.thresholds.detach_()

    def accuracy(self, batch_size):
        """

        Parameters
        ----------
        batch_size: int

        Returns
        -------
        acc: float
            Share of all triplets (true and negatively sampled ones) that where correctly\
            classified using the thresholds learned from the validation set.

        """
        if not self.evaluated:
            self.evaluate(batch_size)

        r_idx = self.kg_test.relations

        neg_heads, neg_tails = self.sampler.corrupt_kg(batch_size, self.use_cuda, which='test')
        scores = self.get_scores(self.kg_test.head_idx, self.kg_test.tail_idx, r_idx, batch_size)
        neg_scores = self.get_scores(neg_heads, neg_tails, r_idx, batch_size)

        if self.use_cuda:
            self.thresholds = self.thresholds.cuda()

        scores = (scores > self.thresholds[r_idx]).byte()
        neg_scores = (neg_scores < self.thresholds[r_idx]).byte()

        return (scores.sum().item() + neg_scores.sum().item()) / (2 * self.kg_test.n_sample)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import empty, tensor, cat
from torch.utils.data import Dataset

from torchkge.utils import get_dictionaries
from torchkge.exceptions import SizeMismatchError

from tqdm import tqdm
from collections import defaultdict


class SmallKG(Dataset):
    """Minimalist version of a knowledge graph. Built with tensors of heads, tails and relations.

    """
    def __init__(self, heads, tails, relations):
        assert heads.shape == tails.shape == relations.shape
        self.heads = heads
        self.tails = tails
        self.relations = relations
        self.length = heads.shape[0]

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        return self.heads[item].item(), self.tails[item].item(), self.relations[item].item()


class KnowledgeGraph(Dataset):
    """Knowledge graph representation.

        Parameters
        ----------
        df: pandas.DataFrame
            Data frame containing three columns [from, to, rel].
        ent2ix: dict, optional
            Dictionary mapping entity labels to their integer key.
        rel2ix: dict, optional
            Dictionary mapping relation labels to their integer key.

        Attributes
        ----------
        ent2ix: dict
            Dictionary mapping entity labels to their integer key.
        rel2ix: dict
            Dictionary mapping relation labels to their integer key.
        n_ent: int
            Number of distinct entities in the data set.
        n_rel: int
            Number of distinct entities in the data set.
        n_sample: int
            Number of samples in the data set. A sample is a fact: a triplet (h, r, l).
        head_idx: torch.Tensor, dtype = long, shape = (n_sample)
            List of the int key of heads for each sample (fact).
        tail_idx: torch.Tensor, dtype = long, shape = (n_sample)
            List of the int key of tails for each sample (facts).
        relations: torch.Tensor, dtype = long, shape = (n_sample)
            List of the int key of relations for each sample (facts).

    """

    def __init__(self, df=None, kg=None,
                 ent2ix=None, rel2ix=None, dict_of_heads=None, dict_of_tails=None):

        if ent2ix is None:
            self.ent2ix = get_dictionaries(df, ent=True)
        else:
            self.ent2ix = ent2ix

        if rel2ix is None:
            self.rel2ix = get_dictionaries(df, ent=False)
        else:
            self.rel2ix = rel2ix

        self.n_ent = max(self.ent2ix.values()) + 1
        self.n_rel = max(self.rel2ix.values()) + 1

        if df is not None:
            assert kg is None
            self.df = df
            self.n_sample = len(df)
            self.head_idx = tensor(df['from'].map(self.ent2ix).values).long()
            self.tail_idx = tensor(df['to'].map(self.ent2ix).values).long()
            self.relations = tensor(df['rel'].map(self.rel2ix).values).long()
        else:
            assert kg is not None
            self.df = kg['df']
            self.n_sample = kg['heads'].shape[0]
            self.head_idx = kg['heads']
            self.tail_idx = kg['tails']
            self.relations = kg['relations']

        if dict_of_heads is None or dict_of_tails is None:
            self.dict_of_heads = defaultdict(set)
            self.dict_of_tails = defaultdict(set)
            print('Evaluating dictionaries of possible heads and tails for relations.')
            self.evaluate_dicts()

        else:
            self.dict_of_heads = dict_of_heads
            self.dict_of_tails = dict_of_tails

    def __len__(self):
        return self.n_sample

    def __getitem__(self, item):
        return self.head_idx[item].item(), self.tail_idx[item].item(), self.relations[item].item()

    def split_kg(self, share=0.8, sizes=None, validation=False):
        """Split the knowledge graph into train and test.

        Parameters
        ----------
        share: float
            Percentage to allocate to train set.
        sizes: tuple
            Tuple of ints of length 2 or 3. If len(sizes) == 2, then the first sizes[0] values of\
            the knowledge graph will be used as training set and the rest as test set.\
            If len(sizes) == 3, the first sizes[0] values of the knowledge graph will be used as\
            training set, the following sizes[1] as validation set and the last sizes[2] as testing\
            set.
        validation: bool
            Indicate if a validation set should be produced along with train and test sets.

        Returns
        -------
        train_kg: torchkge.data.KnowledgeGraph
        val_kg: torchkge.data.KnowledgeGraph (optional)
        test_kg: torchkge.data.KnowledgeGraph

        """

        if sizes is not None:
            try:
                if len(sizes) == 3:
                    assert (sizes[0] + sizes[1] + sizes[2] == self.n_sample)
                elif len(sizes) == 2:
                    assert (sizes[0] + sizes[1] == self.n_sample)
                else:
                    raise SizeMismatchError('Tuple `sizes` should be of length 2 or 3.')
            except AssertionError:
                raise SizeMismatchError('Tuple `sizes` should sum up to the number of facts in the '
                                        'knowledge graph.')

        if ((sizes is not None) and (len(sizes) == 3)) or ((sizes is None) and validation):
            if (sizes is None) and validation:
                samp = empty(self.head_idx.shape).uniform_()
                mask_tr = (samp < share)
                mask_val = (samp > share) & (samp < (1 + share) / 2)
                mask_te = ~(mask_tr | mask_val)
            else:
                mask_tr = cat([tensor([1 for _ in range(sizes[0])]),
                               tensor([0 for _ in range(sizes[1] + sizes[2])])]).byte()
                mask_val = cat([tensor([0 for _ in range(sizes[0])]),
                                tensor([1 for _ in range(sizes[1])]),
                                tensor([0 for _ in range(sizes[2])])]).byte()
                mask_te = ~(mask_tr | mask_val)

            return KnowledgeGraph(
                kg={'heads': self.head_idx[mask_tr], 'tails': self.tail_idx[mask_tr],
                    'relations': self.relations[mask_tr], 'df': self.df},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[mask_val], 'tails': self.tail_idx[mask_val],
                    'relations': self.relations[mask_val], 'df': self.df},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[mask_te], 'tails': self.tail_idx[mask_te],
                    'relations': self.relations[mask_te], 'df': self.df},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails)
        else:
            assert (((sizes is not None) and len(sizes) == 2) or
                    ((sizes is None) and not validation))
            if sizes is None:
                mask_tr = (empty(self.head_idx.shape).uniform_() < share)
            else:
                mask_tr = cat([tensor([1 for _ in range(sizes[0])]),
                               tensor([0 for _ in range(sizes[1])])]).byte()
            return KnowledgeGraph(
                kg={'heads': self.head_idx[mask_tr], 'tails': self.tail_idx[mask_tr],
                    'relations': self.relations[mask_tr], 'df': self.df},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[~mask_tr], 'tails': self.tail_idx[~mask_tr],
                    'relations': self.relations[~mask_tr], 'df': self.df},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails)

    def evaluate_dicts(self):
        """Evaluate dicts of possible alternatives to an entity in a fact that still gives a true\
        fact in the entire knowledge graph.

        """
        for i in tqdm(range(self.n_sample)):
            self.dict_of_heads[(self.tail_idx[i].item(),
                                self.relations[i].item())].add(self.head_idx[i].item())
            self.dict_of_tails[(self.head_idx[i].item(),
                                self.relations[i].item())].add(self.tail_idx[i].item())



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""
from torch import tensor, bernoulli, randint, ones, rand, cat
from torch.utils.data import DataLoader

from torchkge.utils import get_bern_probs, fill_in_dicts
from torchkge.exceptions import NotYetImplementedError


class NegativeSampler:
    """This is an interface for negative samplers in general.

    Parameters
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.

    Attributes
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.
    n_ent: int
        Number of entities in the entire knowledge graph. This is the same in `kg`, `kg_val`\
         and `kg_test`.
    n_sample: int
        Number of triplets in `kg`.
    n_sample_val: in
        Number of triplets in `kg_val`.
    n_sample_test: int
        Number of triples in `kg_test`.

    """
    def __init__(self, kg, kg_val=None, kg_test=None):
        self.kg = kg
        self.n_ent = kg.n_ent
        self.n_sample = kg.n_sample

        self.kg_val = kg_val
        self.kg_test = kg_test

        if kg_val is None:
            self.n_sample_val = 0
        else:
            self.n_sample_val = kg_val.n_sample

        if kg_test is None:
            self.n_sample_test = 0
        else:
            self.n_sample_test = kg_test.n_sample

    def corrupt_batch(self, heads, tails, relations):
        raise NotYetImplementedError('NegativeSampler is just an interface, please consider using '
                                     'a child class where this is implemented.')

    def corrupt_kg(self, batch_size, use_cuda, which='main'):
        """Corrupt an entire knowledge graph using a dataloader and by calling `corrupt_batch`
        methods.

        Parameters
        ----------
        batch_size: int
            Size of the batches used in the dataloader.
        use_cuda: bool
            Indicate whether to use cuda or not
        which: str
            Indicate which graph should be corrupted. Possible values are :\
            * 'main': attribute self.kg is corrupted
            * 'train': attribute self.kg is corrupted
            * 'val': attribute self.kg_val is corrupted. In this case this attribute should have\
            been initialized.
            * 'test': attribute self.kg_test is corrupted. In this case this attribute should have\
            been initialized.

        Returns
        -------
        neg_heads: torch.Tensor, dtype = long, shape = (n_samples)
            Tensor containing the integer key of negatively sampled heads of the relations\
            in the graph designated by `which`.
        neg_tails: torch.Tensor, dtype = long, shape = (n_samples)
            Tensor containing the integer key of negatively sampled tails of the relations\
            in the graph designated by `which`.
        """
        assert which in ['main', 'train', 'test', 'val']
        if which == 'val':
            assert self.n_sample_val > 0
        if which == 'test':
            assert self.n_sample_test > 0

        if which == 'val':
            dataloader = DataLoader(self.kg_val, batch_size=batch_size, shuffle=False,
                                    pin_memory=use_cuda)
        elif which == 'test':
            dataloader = DataLoader(self.kg_test, batch_size=batch_size, shuffle=False,
                                    pin_memory=use_cuda)
        else:
            dataloader = DataLoader(self.kg, batch_size=batch_size, shuffle=False,
                                    pin_memory=use_cuda)

        corr_heads, corr_tails = [], []

        for i, batch in enumerate(dataloader):

            heads, tails, rels = batch[0], batch[1], batch[2]
            if heads.is_pinned():
                heads, tails, rels = heads.cuda(), tails.cuda(), rels.cuda()

            neg_heads, neg_tails = self.corrupt_batch(heads, tails, rels)

            corr_heads.append(neg_heads)
            corr_tails.append(neg_tails)

        if use_cuda:
            return cat(corr_heads).long().cpu(), cat(corr_tails).long().cpu()
        else:
            return cat(corr_heads).long(), cat(corr_tails).long()


class UniformNegativeSampler(NegativeSampler):
    """Uniform negative sampler as presented in 2013 paper by Bordes et al.. Either the head or\
    the tail of a triplet is replaced by another entity at random. The choice of head/tail is\
    uniform.

    References
    ----------
    * Antoine Bordes, Nicolas Usunier, Alberto Garcia-Duran, Jason Weston, and Oksana Yakhnenko.
      Translating Embeddings for Modeling Multi-relational Data.
      In Advances in Neural Information Processing Systems 26, pages 2787–2795, 2013.
      https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data

    Parameters
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.

    Attributes
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.
    n_ent: int
        Number of entities in the entire knowledge graph. This is the same in `kg`, `kg_val`\
         and `kg_test`.
    n_sample: int
        Number of triplets in `kg`.
    n_sample_val: in
        Number of triplets in `kg_val`.
    n_sample_test: int
        Number of triples in `kg_test`.

    """
    def __init__(self, kg, kg_val=None, kg_test=None):
        super().__init__(kg, kg_val, kg_test)

    def corrupt_batch(self, heads, tails, relations=None):
        """For each golden triplet, produce a corrupted one not different from any other golden\
        triplet. If `heads` and `tails` are cuda objects , then the returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of heads of the relations in the current batch.
        tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of tails of the relations in the current batch.
        relations: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of relations in the current batch. This is optional\
            here and mainly present because of the interface with other NegativeSampler objects.

        Returns
        -------
        neg_heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled heads of the relations\
            in the current batch.
        neg_tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled tails of the relations\
            in the current batch.
        """
        use_cuda = heads.is_cuda
        assert (use_cuda == tails.is_cuda)
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'

        batch_size = heads.shape[0]
        neg_heads, neg_tails = heads.clone(), tails.clone()

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(ones(size=(batch_size,), device=device) / 2).double()
        n_heads_corrupted = int(mask.sum().item())
        neg_heads[mask == 1] = randint(1, self.n_ent, (n_heads_corrupted,), device=device)
        neg_tails[mask == 0] = randint(1, self.n_ent, (batch_size - n_heads_corrupted,),
                                       device=device)

        return neg_heads.long(), neg_tails.long()


class BernoulliNegativeSampler(NegativeSampler):
    """Bernoulli negative sampler as presented in 2014 paper by Wang et al.. Either the head or\
    the tail of a triplet is replaced by another entity at random. The choice of head/tail is done\
    using probabilities taking into account profiles of the relations. See the paper for more\
    details.

    References
    ----------
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      Knowledge Graph Embedding by Translating on Hyperplanes.
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.
      https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531

    Parameters
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.

    Attributes
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.
    n_ent: int
        Number of entities in the entire knowledge graph. This is the same in `kg`, `kg_val`\
         and `kg_test`.
    n_sample: int
        Number of triplets in `kg`.
    n_sample_val: in
        Number of triplets in `kg_val`.
    n_sample_test: int
        Number of triples in `kg_test`.
    bern_probs: torch.Tensor, dtype = float, shape = (kg.n_rel)
        Bernoulli sampling probabilities. See paper for more details.

    """
    def __init__(self, kg, kg_val=None, kg_test=None):
        super().__init__(kg, kg_val, kg_test)
        self.bern_probs = self.evaluate_probabilities()

    def evaluate_probabilities(self):
        """Evaluate the Bernoulli probabilities for negative sampling as in the TransH original\
        paper by Wang et al. (2014) https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531.\
        Currently it is done using a pandas DataFrame. This should change as soon as the authors\
        find an efficient way to group-by in torch. TODO

        """
        bern_probs = get_bern_probs(self.kg.df)
        assert len(bern_probs) == self.kg.n_rel
        bern_probs = {self.kg.rel2ix[rel]: bern_probs[rel] for rel in bern_probs.keys()}
        return tensor([bern_probs[k] for k in sorted(bern_probs.keys())]).float()

    def corrupt_batch(self, heads, tails, relations):
        """For each golden triplet, produce a corrupted one different from any other golden\
        triplet. If `heads` and `tails` are cuda objects , then the returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of heads of the relations in the current batch.
        tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of tails of the relations in the current batch.
        relations: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of relations in the current batch.

        Returns
        -------
        neg_heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled heads of the relations\
            in the current batch.
        neg_tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled tails of the relations\
            in the current batch.
        """
        use_cuda = heads.is_cuda
        assert (use_cuda == tails.is_cuda)
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'

        batch_size = heads.shape[0]
        neg_heads, neg_tails = heads.clone(), tails.clone()

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(self.bern_probs[relations]).double()
        n_heads_corrupted = int(mask.sum().item())
        neg_heads[mask == 1] = randint(1, self.n_ent, (n_heads_corrupted,), device=device)
        neg_tails[mask == 0] = randint(1, self.n_ent, (batch_size - n_heads_corrupted,),
                                       device=device)

        return neg_heads.long(), neg_tails.long()


class PositionalNegativeSampler(BernoulliNegativeSampler):
    """Positional negative sampler as presented in 2011 paper by Socher et al.. Either the head or\
    the tail of a triplet is replaced by another entity chosen among entities that have already\
    appeared at the same place in a triplet (involving the same relation). It is not clear in the\
    paper how the choice of head/tail is done. We chose to use Bernoulli sampling as in 2014 paper\
    by Wang et al. as we believe it serves the same purpose as the original paper.

    References
    ----------
    * Richard Socher, Danqi Chen, Christopher D Manning, and Andrew Ng.
      Reasoning With Neural Tensor Networks for Knowledge Base Completion.
      In Advances in Neural Information Processing Systems 26, pages 926–934., 2013.
      https://nlp.stanford.edu/pubs/SocherChenManningNg_NIPS2013.pdf
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      Knowledge Graph Embedding by Translating on Hyperplanes.
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.
      https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531

    Parameters
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.

    Attributes
    ----------
    kg: torchkge.data.KnowledgeGraph.KnowledgeGraph
        Main knowledge graph (usually training one).
    kg_val: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Validation knowledge graph.
    kg_test: torchkge.data.KnowledgeGraph.KnowledgeGraph (optional)
        Test knowledge graph.
    n_ent: int
        Number of entities in the entire knowledge graph. This is the same in `kg`, `kg_val`\
         and `kg_test`.
    n_sample: int
        Number of triplets in `kg`.
    n_sample_val: in
        Number of triplets in `kg_val`.
    n_sample_test: int
        Number of triples in `kg_test`.
    bern_probs: torch.Tensor, dtype = float, shape = (kg.n_rel)
        Bernoulli sampling probabilities. See paper for more details.
    possible_heads: dict
        keys : relations, values : list of possible heads for each relation.
    possible_tails: dict
        keys : relations, values : list of possible tails for each relation.
    n_poss_heads: list
        List of number of possible heads for each relation.
    n_poss_tails: list
        List of number of possible tails for each relation.

    """
    def __init__(self, kg, kg_val=None, kg_test=None):
        super().__init__(kg, kg_val, kg_test)
        self.possible_heads, self.possible_tails, self.n_poss_heads, self.n_poss_tails = self.find_possibilities()

    def find_possibilities(self):
        """For each relation of the knowledge graph (and possibly the validation graph but not the\
        test graph) find all the possible heads and tails in the sens of Wang et al., e.g. all\
        entities that occupy once this position in another triplet.

        Returns
        -------
        possible_heads: dict
            keys : relation index, values : list of possible heads
        possible tails: dict
            keys : relation index, values : list of possible tails
        n_poss_heads: torch.Tensor, dtype = long, shape = (n_relations)
            Number of possible heads for each relation.
        n_poss_tails: torch.Tensor, dtype = long, shape = (n_relations)
            Number of possible tails for each relation.

        """
        possible_heads, possible_tails = fill_in_dicts(self.kg)

        if self.n_sample_val > 0:
            possible_heads, possible_tails = fill_in_dicts(self.kg_val,
                                                           possible_heads, possible_tails)

        possible_heads = dict(possible_heads)
        possible_tails = dict(possible_tails)

        n_poss_heads = []
        n_poss_tails = []

        for i in range(self.kg.n_rel):
            n_poss_heads.append(len(possible_heads[i]))
            n_poss_tails.append(len(possible_tails[i]))
            possible_heads[i] = list(possible_heads[i])
            possible_tails[i] = list(possible_tails[i])

        n_poss_heads = tensor(n_poss_heads)
        n_poss_tails = tensor(n_poss_tails)

        return possible_heads, possible_tails, n_poss_heads, n_poss_tails

    def corrupt_batch(self, heads, tails, relations):
        """For each golden triplet, produce a corrupted one not different from any other golden\
        triplet. If `heads` and `tails` are cuda objects , then the returned tensors are on the GPU.

        Parameters
        ----------
        heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of heads of the relations in the current batch.
        tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of tails of the relations in the current batch.
        relations: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of relations in the current batch. This is optional\
            here and mainly present because of the interface with other NegativeSampler objects.

        Returns
        -------
        neg_heads: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled heads of the relations\
            in the current batch.
        neg_tails: torch.Tensor, dtype = long, shape = (batch_size)
            Tensor containing the integer key of negatively sampled tails of the relations\
            in the current batch.
        """
        use_cuda = heads.is_cuda
        assert (use_cuda == tails.is_cuda)
        if use_cuda:
            device = 'cuda'
        else:
            device = 'cpu'

        batch_size = heads.shape[0]
        neg_heads, neg_tails = heads.clone(), tails.clone()

        # Randomly choose which samples will have head/tail corrupted
        mask = bernoulli(self.bern_probs[relations]).double()
        n_heads_corrupted = int(mask.sum().item())

        # Get the number of possible entities for head and tail
        n_poss_heads = self.n_poss_heads[relations[mask == 1]]
        n_poss_tails = self.n_poss_tails[relations[mask == 0]]

        assert n_poss_heads.shape[0] == n_heads_corrupted
        assert n_poss_tails.shape[0] == batch_size - n_heads_corrupted

        # Choose a rank of an entity in the list of possible entities
        choice_heads = (n_poss_heads.float() * rand((n_heads_corrupted,))).floor().long()
        choice_tails = (
                n_poss_tails.float() * rand((batch_size - n_heads_corrupted,))).floor().long()

        corr = []
        rels = relations[mask == 1]
        for i in range(n_heads_corrupted):
            r = rels[i].item()
            corr.append(self.possible_heads[r][choice_heads[i].item()])
        neg_heads[mask == 1] = tensor(corr, device=device)

        corr = []
        rels = relations[mask == 0]
        for i in range(batch_size - n_heads_corrupted):
            r = rels[i].item()
            corr.append(self.possible_tails[r][choice_tails[i].item()])
        neg_tails[mask == 0] = tensor(corr, device=device)

        return neg_heads.long(), neg_tails.long()



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
aboschin@enst.fr
"""

from torch import Tensor, tensor

from .operations import concatenate_diff_sizes


def get_dictionaries(df, ent=True):
    """Build entities or relations dictionaries.

    Parameters
    ----------
    df: pandas Dataframe
        Data frame containing three columns [from, to, rel].
    ent: bool
        if True then ent2ix is returned, if False then rel2ix is returned.

    Returns
    -------
    dict: dictionary
        Either ent2ix or rel2ix.

    """
    if ent:
        tmp = list(set(df['from'].unique()).union(set(df['to'].unique())))
        return {ent: i for i, ent in enumerate(tmp)}
    else:
        tmp = list(df['rel'].unique())
        return {rel: i for i, rel in enumerate(tmp)}


def lists_from_dicts(dictionary, entities, relations, targets, cuda):
    """

    Parameters
    ----------
    dictionary: dict
        keys: (ent, rel), values: list of entities
    entities: torch.Tensor, dtype = long, shape = (batch_size)
        Heads (resp. tails) of facts.
    relations: torch.Tensor, dtype = long, shape = (batch_size)
        Relations of facts
    targets: torch.Tensor, dtype = long, shape = (batch_size)
        Tails (resp. heads) of facts.
    cuda: bool
        If True, result is returned as CUDA tensor.

    Returns
    -------
    result: torch.Tensor, dtype = long, shape = (k)
        k is the largest number of possible alternative to the target in a fact. This tensor\
        contains for each line (fact) the list of possible alternatives to the target. If there\
        are no alternatives, then the line is full of -1.

    """
    result = Tensor().long()

    if entities.is_cuda:
        result = result.cuda()

    for i in range(entities.shape[0]):
        current = dictionary[(entities[i].item(), relations[i].item())]
        current.remove(targets[i])
        if len(current) == 0:
            current.append(-1)
        current = tensor(current).long().view(1, -1)
        if entities.is_cuda:
            current = current.cuda()
        result = concatenate_diff_sizes(result, current)
    if cuda:
        return result.cuda()
    else:
        return result.cpu()


def get_tph(df):
    df = df.groupby(['from', 'rel']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_hpt(df):
    df = df.groupby(['rel', 'to']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_bern_probs(df):
    """Evaluate the Bernoulli probabilities for negative sampling as in the TransH original\
    paper by Wang et al. (2014) https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531.

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe from which the torchkge.data.KnowledgeGraph object is built.

    Returns
    -------
    tph: dict
        keys: relations as they appear in the pandas dataframe df, values: sampling probabilities\
        as described by Wang et al. in their paper.

    """
    hpt = get_hpt(df)
    tph = get_tph(df)

    assert hpt.keys() == tph.keys()

    for k in tph.keys():
        tph[k] = tph[k] / (tph[k] + hpt[k])

    return tph
